import xwrf


def test_xwrf_version_report():
    info = xwrf.show_versions(as_dict=True)
    assert set(info.keys()) == {'sys_info', 'packages_info'}
    assert info['packages_info']['xwrf'] == xwrf.__version__



from __future__ import annotations

import importlib
import typing

import packaging
import pytest


def importorskip(module_name, minversion: str = None) -> typing.Callable:
    try:
        module = importlib.import_module(module_name)

        if minversion and packaging.version.parse(module.__version__) < packaging.version.parse(
            minversion
        ):
            raise ValueError(
                f'{module_name} version {module.__version__} is less than {minversion}'
            )

    except ImportError:
        skip = True
        reason = f'could not import optional module {module_name}'

    except ValueError as err:
        skip = True
        reason = str(err)

    except:
        skip = True
        reason = f'failure in optional module {module_name}'

    else:
        skip = False
        reason = ''

    return pytest.mark.skipif(skip, reason=reason)



"""Provide grid-related functionality specific to WRF datasets.
This submodule contains code reused with modification from Salem under the terms of the BSD
3-Clause License. Salem is Copyright (c) 2014-2021, Fabien Maussion and Salem Development Team All
rights reserved.
"""
from __future__ import annotations  # noqa: F401

from typing import Hashable, Mapping

import numpy as np
import pyproj
import xarray as xr

# Default CRS (lon/lat on WGS84, which is EPSG:4326)
wgs84 = pyproj.CRS(4326)


def _wrf_grid_from_dataset(ds: xr.Dataset) -> Mapping[Hashable, pyproj.CRS | np.ndarray]:
    """Get the WRF projection and dimension coordinates out of the file."""

    # Use standards from a typical WRF file
    cen_lon = ds.CEN_LON
    cen_lat = ds.CEN_LAT
    dx = ds.DX
    dy = ds.DY
    proj_id = ds.MAP_PROJ

    pargs = {
        'x_0': 0,
        'y_0': 0,
        'a': 6370000,
        'b': 6370000,
        'lat_1': ds.TRUELAT1,
        'lat_2': getattr(ds, 'TRUELAT2', ds.TRUELAT1),
        'lat_0': ds.MOAD_CEN_LAT,
        'lon_0': ds.STAND_LON,
        'center_lon': cen_lon,
    }

    if proj_id == 1:
        # Lambert
        pargs['proj'] = 'lcc'
        del pargs['center_lon']
    elif proj_id == 2:
        # Polar stereo
        pargs['proj'] = 'stere'
        pargs['lat_ts'] = pargs['lat_1']
        pargs['lat_0'] = 90.0
        del pargs['lat_1'], pargs['lat_2'], pargs['center_lon']
    elif proj_id == 3:
        # Mercator
        pargs['proj'] = 'merc'
        pargs['lat_ts'] = pargs['lat_1']
        pargs['lon_0'] = pargs['center_lon']
        del pargs['lat_0'], pargs['lat_1'], pargs['lat_2'], pargs['center_lon']
    else:
        raise NotImplementedError(f'WRF proj not implemented yet: {proj_id}')

    # Construct the pyproj CRS (letting errors fail through)
    crs = pyproj.CRS(pargs)

    # Get grid specifications
    trf = pyproj.Transformer.from_crs(wgs84, crs, always_xy=True)
    nx = ds.dims['west_east']
    ny = ds.dims['south_north']
    e, n = trf.transform(cen_lon, cen_lat)
    x0 = -(nx - 1) / 2.0 * dx + e  # DL corner
    y0 = -(ny - 1) / 2.0 * dy + n  # DL corner

    return {
        'crs': crs,
        'south_north': y0 + np.arange(ny) * dy,
        'west_east': x0 + np.arange(nx) * dx,
        'south_north_stag': y0 + (np.arange(ny + 1) - 0.5) * dy,
        'west_east_stag': x0 + (np.arange(nx + 1) - 0.5) * dx,
    }



import pytest

import xwrf


@pytest.fixture(scope='session')
def config():
    return xwrf.config


def test_defaults_exist(config):
    assert config.get('cf_attribute_map.T2.standard_name') is not None



from __future__ import annotations  # noqa: F401

import re
import warnings

import numpy as np
import pandas as pd
import xarray as xr

from .config import config
from .grid import _wrf_grid_from_dataset


def _decode_times(ds: xr.Dataset) -> xr.Dataset:
    """
    Decode the time variable to datetime64.
    """
    try:
        _time = pd.to_datetime(
            ds.Times.data.astype('str'), errors='raise', format='%Y-%m-%d_%H:%M:%S'
        )
    except ValueError:
        _time = pd.to_datetime(
            ds.Times.data.astype('str'), errors='raise', format='%Y-%m-%dT%H:%M:%S.%f'
        )
    ds = ds.assign_coords({'Time': _time})
    ds.Time.attrs = {'long_name': 'Time', 'standard_name': 'time'}
    # make XTIME be consistent with its description
    if 'XTIME' in ds.variables and np.issubdtype(ds.XTIME.dtype, np.datetime64):
        ds['XTIME'].data = (
            ds.XTIME.data
            - pd.to_datetime(
                ds['XTIME'].description, format='minutes since %Y-%m-%d %H:%M:%S'
            ).to_datetime64()
        )
    return ds


def _clean_brackets_from_units(ds: xr.Dataset) -> xr.Dataset:
    """
    Cleans brackets from units attributes
    """
    sep = '\\'
    regex = re.compile(f'[{sep.join(config.get("brackets_to_clean_from_units"))}]')
    for var in ds.variables:
        if 'units' in ds[var].attrs:
            ds[var].attrs['units'] = regex.sub('', ds[var].attrs['units'])
    return ds


def _make_units_pint_friendly(ds: xr.Dataset) -> xr.Dataset:
    """
    Harmonizes awkward WRF units into pint-friendly ones
    """
    ds = _clean_brackets_from_units(ds)
    # We have to invert the mapping from "new_unit -> wrf_units" to "wrf_unit -> new_unit"
    wrf_units_map = {
        v: k for (k, val_list) in config.get('unit_harmonization_map').items() for v in val_list
    }
    for variable in ds.data_vars:
        if ds[variable].attrs.get('units') in wrf_units_map:
            harmonized_unit = wrf_units_map[ds[variable].attrs['units']]
            if harmonized_unit == 'invalid':
                ds[variable].attrs.pop('units', None)
            else:
                ds[variable].attrs['units'] = harmonized_unit
    return ds


def _modify_attrs_to_cf(ds: xr.Dataset) -> xr.Dataset:
    """Modify the attributes of the dataset to comply with CF conventions."""
    # Universal updates
    vars_to_update = set(config.get('cf_attribute_map').keys()).intersection(set(ds.data_vars))
    for variable in vars_to_update:
        ds[variable].attrs.update(config.get(f'cf_attribute_map.{variable}'))

    # Conditional updates (right now just vertical coordinate type)
    hybrid_opt_condition = 'HYBRID_OPT==0' if getattr(ds, 'HYBRID_OPT', 0) == 0 else 'HYBRID_OPT!=0'
    vars_to_update = set(
        config.get(f'conditional_cf_attribute_map.{hybrid_opt_condition}').keys()
    ).intersection(set(ds.data_vars))
    for variable in vars_to_update:
        ds[variable].attrs.update(
            config.get(f'conditional_cf_attribute_map.{hybrid_opt_condition}.{variable}')
        )

    return ds


def _collapse_time_dim(ds: xr.Dataset) -> xr.Dataset:

    # This "time dimension collapsing" assumption is wrong with moving nests
    # and should be applied to static, nested domains.
    lat_lon_coords = set(config.get('latitude_coords') + config.get('longitude_coords'))
    vertical_coords = set(config.get('vertical_coords'))
    coords = set(ds.variables).intersection(lat_lon_coords.union(vertical_coords))
    ds = ds.set_coords(coords)

    for coord in ds.coords:
        data_to_reassign = None
        if coord in lat_lon_coords and ds[coord].ndim == 3:
            data_to_reassign = ds[coord].data[0, :, :]
        elif coord in vertical_coords and ds[coord].ndim == 2:
            data_to_reassign = ds[coord].data[0, :]

        if data_to_reassign is not None:
            attrs, encoding = ds[coord].attrs, ds[coord].encoding
            ds = ds.assign_coords({coord: (ds[coord].dims[1:], data_to_reassign)})
            ds[coord].attrs = attrs
            ds[coord].encoding = encoding

    return ds


def _include_projection_coordinates(ds: xr.Dataset) -> xr.Dataset:
    """Introduce projection dimension coordinate values and CRS."""
    try:
        grid_components = _wrf_grid_from_dataset(ds)
    except KeyError:
        warnings.warn(
            'Unable to create coordinate values and CRS due to insufficient dimensions or '
            'projection metadata.'
        )
        return ds
    horizontal_dims = set(config.get('horizontal_dims')).intersection(set(ds.dims))

    # Include dimension coordinates
    for dim in horizontal_dims:
        ds[dim] = (dim, grid_components[dim], config.get(f'cf_attribute_map.{dim}'))

    # Include CRS
    ds['wrf_projection'] = (tuple(), grid_components['crs'], grid_components['crs'].to_cf())
    for varname in ds.data_vars:
        if any(dim in ds[varname].dims for dim in horizontal_dims):
            ds[varname].attrs['grid_mapping'] = 'wrf_projection'

    return ds


def _assign_coord_to_dim_of_different_name(ds: xr.Dataset) -> xr.Dataset:
    for varname, dim in config.get('assign_coord_to_dim_map').items():
        try:
            ds[dim] = ds[varname]
            del ds[varname]
        except KeyError:
            pass
    return ds


def _rename_dims(ds: xr.Dataset) -> xr.Dataset:
    """Rename dims for more consistent semantics."""
    rename_dim_map = {k: v for k, v in config.get('rename_dim_map').items() if k in ds.dims}
    return ds.rename(rename_dim_map)


def _calc_base_diagnostics(ds: xr.Dataset, drop: bool = True) -> xr.Dataset:
    """Calculate the four basic fields that WRF does not have in physically meaningful form.

    Parameters
    ----------
    dataset : xarray.Dataset
        Dataset representing WRF data opened via normal backend, with chunking.
    drop : bool
        Decide whether to drop the components of origin after creating the diagnostic fields from
        them.

    Notes
    -----
    This operation should be called before destaggering.
    """
    # Potential temperature
    if 'T' in ds.data_vars:
        ds['air_potential_temperature'] = ds['T'] + 300
        ds['air_potential_temperature'].attrs = {
            'units': 'K',
            'standard_name': 'air_potential_temperature',
        }
        if drop:
            del ds['T']

    # Pressure
    if 'P' in ds.data_vars and 'PB' in ds.data_vars:
        ds['air_pressure'] = ds['P'] + ds['PB']
        ds['air_pressure'].attrs = {
            'units': ds['P'].attrs.get('units', 'Pa'),
            'standard_name': 'air_pressure',
        }
        if drop:
            del ds['P'], ds['PB']

    # Geopotential and geopotential height
    if 'PH' in ds.data_vars and 'PHB' in ds.data_vars:
        ds['geopotential'] = ds['PH'] + ds['PHB']
        ds['geopotential'].attrs = {
            'units': 'm**2 s**-2',
            'standard_name': 'geopotential',
            'stagger': ds['PH'].attrs.get('stagger', 'Z'),
        }
        ds['geopotential_height'] = ds['geopotential'] / 9.81
        ds['geopotential_height'].attrs = {
            'units': 'm',
            'standard_name': 'geopotential_height',
            'stagger': ds['PH'].attrs.get('stagger', 'Z'),
        }
        if drop:
            del ds['PH'], ds['PHB']

    return ds



import sys

import pytest
import xarray as xr

from xwrf import tutorial


@pytest.mark.network
class TestLoadDataset:
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.testfile = 'tiny'

    @pytest.fixture
    def simulate_importerror(self, monkeypatch):
        monkeypatch.setitem(sys.modules, 'pooch', None)

    def test_download_from_github(self, tmp_path) -> None:
        cache_dir = tmp_path / tutorial._default_cache_dir_name
        ds = tutorial.open_dataset(self.testfile, cache_dir=cache_dir).load()
        tiny = xr.DataArray(range(5), name='tiny').to_dataset()
        xr.testing.assert_identical(ds, tiny)

    def test_download_from_github_load_without_cache(self, tmp_path, monkeypatch) -> None:
        cache_dir = tmp_path / tutorial._default_cache_dir_name

        ds_nocache = tutorial.open_dataset(self.testfile, cache=False, cache_dir=cache_dir).load()
        ds_cache = tutorial.load_dataset(self.testfile, cache_dir=cache_dir)
        xr.testing.assert_identical(ds_cache, ds_nocache)

    def test_pooch_import_error(self, simulate_importerror):
        with pytest.raises(ImportError):
            tutorial.open_dataset(self.testfile)

    def test_invalid_dataset_key(self):
        with pytest.raises(KeyError):
            tutorial.open_dataset('invalid_dataset_key')



import numpy as np
import pyproj
import pytest

import xwrf
from xwrf.grid import _wrf_grid_from_dataset, wgs84


@pytest.fixture(scope='session', params=['dummy'])
def dummy_dataset(request):
    return xwrf.tutorial.open_dataset(request.param)


@pytest.fixture(scope='session', params=['dummy_salem_parsed'])
def dummy_salem(request):
    return xwrf.tutorial.open_dataset(request.param)


@pytest.fixture(scope='session')
def test_grid(request):
    return xwrf.tutorial.open_dataset(request.param)


@pytest.fixture(scope='session')
def cf_grid_mapping_name(request):
    """A no-op to allow parallel indirect use with open dataset fixture."""
    return request.param


def test_grid_construction_against_salem(dummy_dataset, dummy_salem):
    grid_params = _wrf_grid_from_dataset(dummy_dataset)

    # Projection coordinate values
    np.testing.assert_array_almost_equal(
        grid_params['south_north'], dummy_salem['south_north'].values
    )
    np.testing.assert_array_almost_equal(grid_params['west_east'], dummy_salem['west_east'].values)

    # Projection CRS
    assert grid_params['crs'] == pyproj.CRS(dummy_salem['Q2'].attrs['pyproj_srs'])


@pytest.mark.parametrize(
    'test_grid, cf_grid_mapping_name',
    [
        ('polar_stereographic_1', 'polar_stereographic'),
        ('polar_stereographic_2', 'polar_stereographic'),
        ('lambert_conformal', 'lambert_conformal_conic'),
        ('mercator', 'mercator'),
    ],
    indirect=True,
)
def test_grid_construction_against_own_latlon(test_grid, cf_grid_mapping_name):
    grid_params = _wrf_grid_from_dataset(test_grid)
    trf = pyproj.Transformer.from_crs(grid_params['crs'], wgs84, always_xy=True)
    recalculated = {}
    recalculated['XLONG'], recalculated['XLAT'] = trf.transform(
        *np.meshgrid(grid_params['west_east'], grid_params['south_north'])
    )
    recalculated['XLONG_U'], recalculated['XLAT_U'] = trf.transform(
        *np.meshgrid(grid_params['west_east_stag'], grid_params['south_north'])
    )
    recalculated['XLONG_V'], recalculated['XLAT_V'] = trf.transform(
        *np.meshgrid(grid_params['west_east'], grid_params['south_north_stag'])
    )

    assert grid_params['crs'].to_cf()['grid_mapping_name'] == cf_grid_mapping_name
    for varname, recalculated_values in recalculated.items():
        if varname in test_grid.data_vars:
            np.testing.assert_array_almost_equal(
                recalculated_values,
                test_grid[varname].values[0],
                decimal=2,
                err_msg=f'Computed {varname} does not match with raw output',
            )
        elif '_' not in varname and varname + '_M' in test_grid.data_vars:
            np.testing.assert_array_almost_equal(
                recalculated_values,
                test_grid[varname + '_M'].values[0],
                decimal=2,
                err_msg=f"Computed {varname + '_M'} does not match with raw output",
            )



import metpy
import pytest
import xarray as xr

import xwrf


@pytest.fixture(scope='session')
def sample_dataset(request):
    return xwrf.tutorial.open_dataset(request.param)


@pytest.mark.parametrize('sample_dataset', ['met_em_sample'], indirect=True)
def test_metpy_quantify(sample_dataset):
    ds = sample_dataset.xwrf.postprocess(decode_times=False).metpy.quantify()
    assert str(ds.PRES.data.units) == 'pascal'
    assert str(ds.MAPFAC_U.data.units) == 'dimensionless'


@pytest.mark.parametrize('sample_dataset', ['met_em_sample'], indirect=True)
def test_metpy_parse_cf(sample_dataset):
    ds = sample_dataset.xwrf.postprocess().metpy.parse_cf()
    assert 'metpy_crs' in ds.coords
    assert isinstance(ds.metpy_crs.data.item(), metpy.plots.mapping.CFProjection)
    assert ds['wrf_projection'].item() == ds['metpy_crs'].metpy.pyproj_crs


@pytest.mark.parametrize('sample_dataset', ['lambert_conformal', 'mercator'], indirect=True)
def test_metpy_coordinate_identification(sample_dataset):
    ds = sample_dataset.xwrf.postprocess()
    # U staggered
    xr.testing.assert_identical(ds['U'].metpy.x, ds['x_stag'])
    xr.testing.assert_identical(ds['U'].metpy.y, ds['y'])
    xr.testing.assert_identical(ds['U'].metpy.time, ds['Time'])
    xr.testing.assert_identical(ds['U'].metpy.longitude, ds['XLONG_U'])
    xr.testing.assert_identical(ds['U'].metpy.latitude, ds['XLAT_U'])
    # V staggered
    xr.testing.assert_identical(ds['V'].metpy.x, ds['x'])
    xr.testing.assert_identical(ds['V'].metpy.y, ds['y_stag'])
    xr.testing.assert_identical(ds['V'].metpy.time, ds['Time'])
    xr.testing.assert_identical(ds['V'].metpy.longitude, ds['XLONG_V'])
    xr.testing.assert_identical(ds['V'].metpy.latitude, ds['XLAT_V'])


@pytest.mark.parametrize(
    'sample_dataset, axis_mapping, test_varname',
    [('lambert_conformal', {'x': 3, 'y': 2, 'vertical': 1, 'time': 0}, 'QVAPOR')],
    indirect=['sample_dataset'],
)
def test_metpy_axis_identification(sample_dataset, axis_mapping, test_varname):
    da = sample_dataset.xwrf.postprocess()[test_varname]
    for axis_type, axis_number in axis_mapping.items():
        assert da.metpy.find_axis_number(axis_type) == axis_number
        assert da.metpy.find_axis_name(axis_type) == da.dims[axis_number]



import os

import yaml
from donfig import Config

fn = os.path.join(os.path.dirname(__file__), 'config.yaml')

with open(fn) as f:
    defaults = yaml.safe_load(f)

config = Config('xwrf', defaults=[defaults])
config.ensure_file(fn, comment=False)



import numpy as np
import pyproj
import pytest
import xarray as xr

import xwrf

from . import importorskip


@pytest.fixture(scope='session')
def sample_dataset(request):
    return xwrf.tutorial.open_dataset(request.param)


@pytest.fixture(scope='session')
def sample_dataset_with_kwargs(request):
    return xwrf.tutorial.open_dataset(request.param[0], **request.param[1])


@pytest.mark.parametrize('sample_dataset', ['dummy'], indirect=True)
@pytest.mark.parametrize('variable', ('Q2', 'PSFC'))
def test_cf_attrs_added(sample_dataset, variable):
    dataset = xwrf.postprocess._modify_attrs_to_cf(sample_dataset)
    attrs = dataset[variable].attrs.keys()
    assert 'standard_name' in attrs
    assert 'long_name' in attrs


@pytest.mark.parametrize('sample_dataset', ['dummy_attrs_only'], indirect=True)
@pytest.mark.parametrize('variable', ('THIS_IS_AN_IDEAL_RUN', 'SAVE_TOPO_FROM_REAL'))
def test_remove_invalid_units(sample_dataset, variable):
    dataset = xwrf.postprocess._make_units_pint_friendly(sample_dataset)
    assert 'units' not in dataset[variable].attrs


@pytest.mark.parametrize('sample_dataset', ['met_em_sample'], indirect=True)
@pytest.mark.parametrize('variable', ('OL4', 'GREENFRAC'))
def test_met_em_parsing(sample_dataset, variable):
    dataset = xwrf.postprocess._make_units_pint_friendly(sample_dataset)
    assert dataset[variable].attrs['units'] == '1'


@pytest.mark.parametrize('bad_units', list(xwrf.config.get('unit_harmonization_map').values()))
def test_successful_unit_parsing(bad_units):
    synthetic_dataset = xr.Dataset(coords=dict(x=(['x'], np.random.rand(5))))
    for i, unit in enumerate(bad_units):
        synthetic_dataset[chr(i)] = xr.Variable(
            dims='x', data=np.random.rand(5), attrs={'units': unit}
        )
    xwrf.postprocess._make_units_pint_friendly(synthetic_dataset)


@pytest.mark.parametrize('sample_dataset', ['dummy'], indirect=True)
def test_include_projection_coordinates(sample_dataset):
    dataset = xwrf.postprocess._include_projection_coordinates(sample_dataset)
    assert dataset['south_north'].attrs['axis'] == 'Y'
    assert dataset['west_east'].attrs['axis'] == 'X'
    assert isinstance(dataset['wrf_projection'].item(), pyproj.CRS)
    assert dataset['Q2'].attrs['grid_mapping'] == 'wrf_projection'


@pytest.mark.parametrize(
    'sample_dataset', set(xwrf.tutorial.sample_datasets.keys()) - {'tiny'}, indirect=True
)
def test_grid_mapping_is_in_all_vars(sample_dataset):
    dataset = xwrf.postprocess._include_projection_coordinates(sample_dataset)
    horizontal_dims = set(xwrf.config.get('horizontal_dims')).intersection(set(dataset.dims))
    for var in dataset.data_vars:
        if any(dim in dataset[var].dims for dim in horizontal_dims):
            assert dataset[var].attrs['grid_mapping'] == 'wrf_projection'


@importorskip('xgcm')
@pytest.mark.parametrize('sample_dataset', ['lambert_conformal', 'mercator'], indirect=True)
def test_include_projection_coordinates_with_xgcm(sample_dataset):
    from xgcm import Grid

    dataset = xwrf.postprocess._include_projection_coordinates(sample_dataset)
    grid = Grid(dataset)

    assert grid.axes['Y'].coords['center'] == 'south_north'
    assert grid.axes['Y'].coords['outer'] == 'south_north_stag'
    assert grid.axes['X'].coords['center'] == 'west_east'
    assert grid.axes['X'].coords['outer'] == 'west_east_stag'


@pytest.mark.parametrize('sample_dataset', ['dummy_attrs_only'], indirect=True)
def test_warning_on_projection_coordinate_failure(sample_dataset):
    with pytest.warns(UserWarning):
        dataset = xwrf.postprocess._include_projection_coordinates(sample_dataset)
    assert dataset is sample_dataset


@pytest.mark.parametrize('sample_dataset', ['dummy'], indirect=True)
def test_rename_dims(sample_dataset):
    dataset = xwrf.postprocess._rename_dims(sample_dataset)
    assert {'x', 'y'}.intersection(set(dataset.dims))
    assert not {'south_north', 'west_east'}.intersection(set(dataset.dims))


@pytest.mark.parametrize('times', [[b'2005-08-28_12:00:00'], [b'2005-08-28T12:00:00']])
def test_decode_times(times):
    ds = xr.Dataset({'Times': times})
    dsa = xwrf.postprocess._decode_times(ds)
    assert np.issubdtype(dsa['Time'].dtype, np.datetime64)
    assert dsa['Time'].attrs['long_name'] == 'Time'
    assert dsa['Time'].attrs['standard_name'] == 'time'


@pytest.mark.parametrize(
    'sample_dataset_with_kwargs,xtime_dtype',
    [
        (('mercator', {'decode_times': True}), np.timedelta64),
        (('mercator', {'decode_times': False}), np.float32),
    ],
    indirect=['sample_dataset_with_kwargs'],
)
def test_xtime_handling(sample_dataset_with_kwargs, xtime_dtype):
    dataset = xwrf.postprocess._decode_times(sample_dataset_with_kwargs)
    assert np.issubdtype(dataset.XTIME.dtype, xtime_dtype)


@pytest.mark.parametrize('sample_dataset', ['lambert_conformal'], indirect=True)
def test_assign_coord_to_dim_of_different_name(sample_dataset):
    dataset = sample_dataset.pipe(xwrf.postprocess._collapse_time_dim).pipe(
        xwrf.postprocess._assign_coord_to_dim_of_different_name
    )
    assert 'ZNU' not in dataset.dims
    assert 'ZNW' not in dataset.dims
    assert 'bottom_top' in dataset.dims
    assert 'bottom_top_stag' in dataset.dims
    sample_dataset_w_collapsed_time = sample_dataset.pipe(xwrf.postprocess._collapse_time_dim)
    np.testing.assert_equal(sample_dataset_w_collapsed_time['ZNU'].data, dataset['bottom_top'].data)
    np.testing.assert_equal(
        sample_dataset_w_collapsed_time['ZNW'].data, dataset['bottom_top_stag'].data
    )


@pytest.mark.parametrize('sample_dataset', ['met_em_sample'], indirect=True)
def test_assign_coord_to_dim_of_different_name_keyerror(sample_dataset):
    ds = sample_dataset.pipe(xwrf.postprocess._collapse_time_dim)
    dataset = ds.pipe(xwrf.postprocess._assign_coord_to_dim_of_different_name)
    xr.testing.assert_equal(ds, dataset)


@pytest.mark.parametrize('sample_dataset', ['lambert_conformal'], indirect=True)
@pytest.mark.parametrize('variable,bracket', (('QNRAIN', '('), ('NOAHRES', '{')))
def test_clean_brackets_from_units(sample_dataset, variable, bracket):
    ds = sample_dataset.pipe(xwrf.postprocess._clean_brackets_from_units)
    assert bracket not in ds[variable].attrs['units']


@pytest.mark.parametrize('sample_dataset', ['lambert_conformal'], indirect=True)
def test_calc_base_diagnostics(sample_dataset):
    subset = sample_dataset[['T', 'P', 'PB', 'PH', 'PHB']].isel(Time=0).load()
    ds_dropped = subset.copy().pipe(xwrf.postprocess._calc_base_diagnostics)
    ds_undropped = subset.copy().pipe(xwrf.postprocess._calc_base_diagnostics, drop=False)

    for ds in (ds_dropped, ds_undropped):
        # Potential temperature
        np.testing.assert_allclose(ds['air_potential_temperature'].max().item(), 508.78415)
        np.testing.assert_allclose(ds['air_potential_temperature'].min().item(), 270.265106)
        assert ds['air_potential_temperature'].attrs['units'] == 'K'
        assert ds['air_potential_temperature'].attrs['standard_name'] == 'air_potential_temperature'

        # Air pressure
        np.testing.assert_allclose(ds['air_pressure'].max().item(), 101824.89)
        np.testing.assert_allclose(ds['air_pressure'].min().item(), 5269.97)
        assert ds['air_pressure'].attrs['units'] == 'Pa'
        assert ds['air_pressure'].attrs['standard_name'] == 'air_pressure'

        # Geopotential
        np.testing.assert_allclose(ds['geopotential'].max().item(), 196788.02)
        np.testing.assert_allclose(ds['geopotential'].min().item(), 0.0)
        assert ds['geopotential'].attrs['units'] == 'm**2 s**-2'
        assert ds['geopotential'].attrs['standard_name'] == 'geopotential'

        # Geopotential height
        np.testing.assert_allclose(ds['geopotential_height'].max().item(), 20059.941)
        np.testing.assert_allclose(ds['geopotential_height'].min().item(), 0.0)
        assert ds['geopotential_height'].attrs['units'] == 'm'
        assert ds['geopotential_height'].attrs['standard_name'] == 'geopotential_height'

    # Check dropped or not
    assert 'T' not in ds_dropped
    assert 'P' not in ds_dropped
    assert 'PB' not in ds_dropped
    assert 'PH' not in ds_dropped
    assert 'PHB' not in ds_dropped
    assert 'T' in ds_undropped
    assert 'P' in ds_undropped
    assert 'PB' in ds_undropped
    assert 'PH' in ds_undropped
    assert 'PHB' in ds_undropped


@pytest.mark.parametrize(
    'sample_dataset', ['dummy', 'polar_stereographic_1', 'tiny', 'met_em_sample'], indirect=True
)
def test_calc_base_diagnostics_skipping(sample_dataset):
    ds = sample_dataset.pipe(xwrf.postprocess._calc_base_diagnostics)
    xr.testing.assert_identical(sample_dataset, ds)



import datetime
import xwrf
import sphinx_autosummary_accessors

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'myst_nb',
    'sphinxext.opengraph',
    'sphinx_copybutton',
    'sphinxcontrib.autodoc_pydantic',
    'sphinx_autosummary_accessors',
    'sphinx_design',
]

# Autosummary pages will be generated by sphinx-autogen instead of sphinx-build
autosummary_generate = True

autodoc_typehints = 'description'
autodoc_typehints_description_target = 'documented'
autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'private-members': True,
}

autodoc_member_order = 'groupwise'
autodoc_pydantic_model_show_json = True
autodoc_pydantic_settings_show_json = False

# Napoleon configurations

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False

numpydoc_show_class_members = False
numpydoc_validation_checks = {'all'}


# MyST config
myst_enable_extensions = ['amsmath', 'colon_fence', 'deflist', 'html_image']
myst_url_schemes = ['http', 'https', 'mailto']

# sphinx-copybutton configurations
copybutton_prompt_text = r'>>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: '
copybutton_prompt_is_regexp = True

extlinks = {
    'issue': ('https://github.com/xarray-contrib/xwrf/issues/%s', 'GH#'),
    'pr': ('https://github.com/xarray-contrib/xwrf/pull/%s', 'GH#'),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates', sphinx_autosummary_accessors.templates_path]

# The master toctree document.
master_doc = 'index'

# General information about the project.
current_year = datetime.datetime.now().year
project = 'xWRF'
author = 'xWRF Developers'
copyright = f'2021-{current_year}, {author}'

# The short X.Y version.
version = xwrf.__version__.split('+')[0]
# The full version, including alpha/beta/rc tags.
release = xwrf.__version__


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '**.ipynb_checkpoints', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'furo'
html_title = ''

html_context = {
    'github_user': 'xarray-contrib',
    'github_repo': 'xwrf',
    'github_version': 'main',
    'doc_path': 'docs',
}
html_static_path = ['_static']
html_js_files = ['custom.js']
html_theme_options = dict(
    # analytics_id=''  this is configured in rtfd.io
    # canonical_url="",
    light_logo='xwrf_logo_bg_light.svg',
    dark_logo='xwrf_logo_bg_dark.svg',
    sidebar_hide_name=True,
)
html_favicon = '_static/xwrf_favicon.ico'


# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'


# Output file base name for HTML help builder.
htmlhelp_basename = 'xwrfdoc'

# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

latex_documents = [('index', 'xwrf.tex', 'xwrf Documentation', author, 'manual')]
man_pages = [('index', 'xwrf', 'xwrf Documentation', [author], 1)]
texinfo_documents = [
    (
        'index',
        'xwrf',
        'xwrf Documentation',
        author,
        'xwrf',
        'One line description of project.',
        'Miscellaneous',
    )
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'xarray': ('https://xarray.pydata.org/en/stable/', None),
    'metpy': ('https://unidata.github.io/MetPy/latest/', None),
    'xgcm': ('https://xgcm.readthedocs.io/en/latest/', None),
}



from __future__ import annotations

import importlib
import locale
import os
import pathlib
import platform
import struct
import subprocess
import sys


def get_sys_info() -> dict[str, str]:
    """Returns system information as a dict"""

    # get full commit hash
    commit = None
    if pathlib.Path('.git').is_dir() and pathlib.Path('xwrf').is_dir():
        try:
            pipe = subprocess.Popen(
                'git log --format="%H" -n 1'.split(' '),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            so, _ = pipe.communicate()
        except Exception:
            pass
        else:
            if pipe.returncode == 0:
                commit = so
                try:
                    commit = so.decode('utf-8')
                except ValueError:
                    pass
                commit = commit.strip().strip('"')

    blob = [('xWRF commit', commit)]
    try:
        (sysname, _nodename, release, _version, machine, processor) = platform.uname()
        blob.extend(
            [
                ('python', sys.version),
                ('python-bits', struct.calcsize('P') * 8),
                ('OS', f'{sysname}'),
                ('OS-release', f'{release}'),
                ('machine', f'{machine}'),
                ('processor', f'{processor}'),
                ('byteorder', f'{sys.byteorder}'),
                ('LC_ALL', f'{os.environ.get("LC_ALL", "None")}'),
                ('LANG', f'{os.environ.get("LANG", "None")}'),
                ('LOCALE', f'{locale.getlocale()}'),
            ]
        )
    except Exception:
        pass

    return blob


def show_versions(as_dict: bool = False) -> None | dict[str, str]:
    """Print out versions of xwrf and its dependencies"""
    packages = [
        ('xwrf', lambda mod: mod.__version__),
        ('xarray', lambda mod: mod.__version__),
        ('numpy', lambda mod: mod.__version__),
        ('pandas', lambda mod: mod.__version__),
        ('matplotlib', lambda mod: mod.__version__),
        ('dask', lambda mod: mod.__version__),
        ('netCDF4', lambda mod: mod.__version__),
        ('donfig', lambda mod: mod.__version__),
        ('xgcm', lambda mod: mod.__version__),
        ('pyproj', lambda mod: mod.__version__),
        ('metpy', lambda mod: mod.__version__),
        ('cf_xarray', lambda mod: mod.__version__),
        ('pint', lambda mod: mod.__version__),
        ('pooch', lambda mod: mod.__version__),
    ]

    packages_blob = {}
    for (modname, version_func) in packages:
        try:
            if modname in sys.modules:
                mod = sys.modules[modname]
            else:
                mod = importlib.import_module(modname)
        except Exception:
            packages_blob[modname] = None

        else:
            try:
                version = version_func(mod)
                packages_blob[modname] = version
            except Exception:
                packages_blob[modname] = 'installed'
    sys_info = get_sys_info()
    packages_info = sorted(packages_blob.items())
    print('\nSystem Information')
    print('------------------')
    print('\n'.join(f'{key:<12}: {value}' for (key, value) in sys_info))
    print('\nInstalled Python Packages')
    print('-------------------------')
    print('\n'.join(f'{modname:<12}: {version}' for (modname, version) in packages_info))

    if as_dict:
        return {'sys_info': sys_info, 'packages_info': packages_blob}



#!/usr/bin/env python3

"""The setup script."""

from setuptools import find_packages, setup

with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

with open('README.md') as f:
    long_description = f.read()


CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Topic :: Scientific/Engineering',
]

setup(
    name='xwrf',
    description='A lightweight interface for working with the Weather Research and Forecasting (WRF) model output in Xarray.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.7',
    maintainer='xWRF Developers',
    classifiers=CLASSIFIERS,
    url='https://xwrf.readthedocs.io',
    project_urls={
        'Documentation': 'https://xwrf.readthedocs.io',
        'Source': 'https://github.com/xarray-contrib/xwrf',
        'Tracker': 'https://github.com/xarray-contrib/xwrf/issues',
        'Discussions/Support': 'https://github.com/xarray-contrib/xwrf/discussions',
    },
    packages=find_packages(exclude=('tests',)),
    include_package_data=True,
    install_requires=install_requires,
    license='Apache 2.0',
    zip_safe=False,
    entry_points={},
    keywords='wrf, xarray',
    use_scm_version={'version_scheme': 'post-release', 'local_scheme': 'dirty-tag'},
)



import pytest

import xwrf

from . import importorskip


@pytest.fixture(scope='session')
def sample_dataset(request):
    return xwrf.tutorial.open_dataset(request.param)


@importorskip('cf_units')
@pytest.mark.parametrize('sample_dataset', ['lambert_conformal'], indirect=True)
@pytest.mark.parametrize(
    'variable,cf_definition_string',
    (
        ('QNRAIN', 'kg-1'),
        ('NOAHRES', 'kg.s-3'),
        ('THIS_IS_AN_IDEAL_RUN', '?'),
        ('SAVE_TOPO_FROM_REAL', '?'),
        ('OL4', '?'),
        ('VAR_SSO', 'm2'),
        ('SEAICE', '?'),
    ),
)
def test_problematic_units_against_udunits_wrfout(sample_dataset, variable, cf_definition_string):
    from cf_units import Unit

    ds = sample_dataset.xwrf.postprocess()
    assert Unit(ds[variable].attrs.get('units', '')).definition == cf_definition_string


@importorskip('cf_units')
@pytest.mark.parametrize('sample_dataset', ['met_em_sample'], indirect=True)
@pytest.mark.parametrize(
    'variable,cf_definition_string',
    (
        ('OL4', '1'),
        ('GREENFRAC', '1'),
        ('PRES', 'm-1.kg.s-2'),
        ('ST', 'K'),
        ('SOILTEMP', 'K'),
        ('RH', '0.01 1'),
    ),
)
def test_problematic_units_against_udunits_met_em(sample_dataset, variable, cf_definition_string):
    from cf_units import Unit

    ds = sample_dataset.xwrf.postprocess()
    assert Unit(ds[variable].attrs.get('units', '')).definition == cf_definition_string



#!/usr/bin/env python3
# flake8: noqa
""" Top-level module. """

from pkg_resources import DistributionNotFound, get_distribution

from . import postprocess, tutorial
from .accessors import WRFDataArrayAccessor, WRFDatasetAccessor
from .config import config
from .version_report import show_versions

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:  # pragma: no cover
    __version__ = 'unknown'  # pragma: no cover



"""
Useful for:
* users learning xwrf
* building tutorials in the documentation.
"""
from __future__ import annotations  # noqa: F401

import os
import pathlib

import xarray as xr

_default_cache_dir_name = 'xwrf_tutorial_data'
base_url = 'https://github.com/xarray-contrib/xwrf-data'
version = 'main'


def _construct_cache_dir(path):
    import pooch

    if isinstance(path, os.PathLike):
        path = os.fspath(path)
    elif path is None:
        path = pooch.os_cache(_default_cache_dir_name)

    return path


sample_datasets = {
    'dummy': 'data/dummy.nc',
    'dummy_attrs_only': 'data/dummy_attrs_only.nc',
    'dummy_salem_parsed': 'data/dummy_salem_parsed.nc',
    'polar_stereographic_1': 'data/geo_em_d01_polarstereo.nc',
    'polar_stereographic_2': 'data/geo_em_d02_polarstereo.nc',
    'lambert_conformal': 'data/lambert_conformal_sample.nc',
    'mercator': 'data/mercator_sample.nc',
    'tiny': 'data/tiny.nc',
    'met_em_sample': 'data/met_em.d01.2005-08-28_12:00:00.nc',
    'wrfout': 'data/wrfout_d03_2018-10-16_00:00:00.nc',
}


# idea borrowed from Seaborn and Xarray
def open_dataset(
    name: str,
    cache: bool = True,
    cache_dir: str | pathlib.Path = None,
    *,
    engine: str = 'netcdf4',
    **kws,
) -> xr.Dataset:
    """
    Open a dataset from the online repository (requires internet).
    If a local copy is found then always use that to avoid network traffic.

    Available datasets:

    * ``"dummy"``
    * ``"dummy_attrs_only"``
    * ``"dummy_salem_parsed"``
    * ``"polar_stereographic_1"``
    * ``"polar_stereographic_2"``
    * ``"lambert_conformal"``
    * ``"mercator"``
    * ``"met_em_sample"``
    * ``"wrfout"``

    Parameters
    ----------
    name : str
        Name of the dataset.
        e.g. 'mercator'
    cache : bool, optional
        If True, then cache data locally for use on subsequent calls
    cache_dir : path-like, optional
        The directory in which to search for and write cached data.
    engine : str, optional
        Name of the backend engine to use.
    **kws : dict, optional
        Additional keyword arguments passed through to the :py:func:`~xarray.open_dataset` function.

    Returns
    -------
    xarray.Dataset
        The dataset.
    """

    try:
        import pooch
    except ImportError as e:
        raise ImportError(
            'tutorial.open_dataset depends on pooch to download and manage datasets.'
            ' To proceed please install pooch using:'
            ' `python -m pip install pooch` or `conda install -c conda-forge pooch`.'
        ) from e

    logger = pooch.get_logger()
    logger.setLevel('WARNING')

    cache_dir = _construct_cache_dir(cache_dir)
    try:
        path = sample_datasets[name]
    except KeyError as exc:
        raise KeyError(
            f'{name} is not a valid dataset name. Valid names include: {list(sample_datasets.keys())}.'
        ) from exc
    url = f'{base_url}/raw/{version}/{path}'

    # retrieve the file
    filepath = pooch.retrieve(url=url, known_hash=None, path=cache_dir)
    ds = xr.open_dataset(filepath, engine=engine, **kws)
    if not cache:
        ds = ds.load()
        pathlib.Path(filepath).unlink()

    return ds


def load_dataset(*args, **kwargs) -> xr.Dataset:
    """
    Open, load into memory, and close a dataset from the online repository
    (requires internet)
    """
    with open_dataset(*args, **kwargs) as ds:
        return ds.load()



from __future__ import annotations  # noqa: F401

import xarray as xr

from .postprocess import (
    _assign_coord_to_dim_of_different_name,
    _calc_base_diagnostics,
    _collapse_time_dim,
    _decode_times,
    _include_projection_coordinates,
    _make_units_pint_friendly,
    _modify_attrs_to_cf,
    _rename_dims,
)


class WRFAccessor:
    """
    Common Dataset and DataArray accessor functionality.
    """

    def __init__(self, xarray_obj: xr.Dataset | xr.DataArray) -> WRFAccessor:
        self.xarray_obj = xarray_obj


@xr.register_dataarray_accessor('xwrf')
class WRFDataArrayAccessor(WRFAccessor):
    """Adds a number of WRF specific methods to xarray.DataArray objects."""


@xr.register_dataset_accessor('xwrf')
class WRFDatasetAccessor(WRFAccessor):
    """Adds a number of WRF specific methods to xarray.Dataset objects."""

    def postprocess(
        self,
        decode_times: bool = True,
        calculate_diagnostic_variables: bool = True,
        drop_diagnostic_variable_components: bool = True,
    ) -> xr.Dataset:
        """
        Postprocess the dataset. This method will perform the following operations:

        - Rename dimensions to match the CF conventions.
        - Rename variables to match the CF conventions.
        - Rename variable attributes to match the CF conventions.
        - Convert units to Pint-friendly units.
        - Decode times.
        - Include projection coordinates.
        - Collapse time dimension.

        Parameters
        ----------
        decode_times : bool, optional
            Decode the string-like wrfout times to xarray-friendly Pandas types. Defaults to True.
        calculate_diagnostic_variables : bool, optional
            Calculate four essential diagnostic variables (potential temperature, air pressure,
            geopotential, and geopotential height) that are otherwise only present in wrfout files
            as split components or dependent upon special adjustments. Defaults to True. If the
            underlying fields on which any of these calculated fields depends is missing, that
            calculated variable is skipped. These will be eagerly evalulated, unless your data has
            been chunked with Dask, in which case these fields will also be Dask arrays.
        drop_diagnostic_variable_components : bool, optional
            Determine whether to drop the underlying fields used to calculate the diagnostic
            variables. Defaults to True.

        Returns
        -------
        xarray.Dataset
            The postprocessed dataset.
        """
        ds = (
            self.xarray_obj.pipe(_modify_attrs_to_cf)
            .pipe(_make_units_pint_friendly)
            .pipe(_collapse_time_dim)
            .pipe(_assign_coord_to_dim_of_different_name)
        )
        if decode_times:
            ds = ds.pipe(_decode_times)
        if calculate_diagnostic_variables:
            ds = ds.pipe(_calc_base_diagnostics, drop=drop_diagnostic_variable_components)

        ds = ds.pipe(_include_projection_coordinates)

        return ds.pipe(_rename_dims)



import pandas as pd
import pytest

import xwrf

from . import importorskip


@importorskip('cf_xarray')
@pytest.mark.parametrize(
    'name, cf_grid_mapping_name', [('lambert_conformal', 'lambert_conformal_conic')]
)
def test_postprocess(name, cf_grid_mapping_name):

    # Verify initial/raw state
    raw_ds = xwrf.tutorial.open_dataset(name)
    assert pd.api.types.is_string_dtype(raw_ds.Times.dtype)
    assert pd.api.types.is_numeric_dtype(raw_ds.Time.dtype)
    assert 'time' not in raw_ds.cf.coordinates
    assert raw_ds.cf.standard_names == {}

    # Postprocess without decoding times
    dsa = raw_ds.xwrf.postprocess(decode_times=False)
    assert pd.api.types.is_numeric_dtype(dsa.Time.dtype)

    # Postprocess
    ds = raw_ds.xwrf.postprocess()

    # Check for time coordinate handling
    assert pd.api.types.is_datetime64_dtype(ds.Time.dtype)
    assert 'time' in ds.cf.coordinates

    # Check for projection handling
    assert ds['wrf_projection'].attrs['grid_mapping_name'] == cf_grid_mapping_name

    # Check for standard name and variable handling
    standard_names = ds.cf.standard_names
    assert 'x' in standard_names['projection_x_coordinate']
    assert 'y' in standard_names['projection_y_coordinate']
    assert 'z' in standard_names['atmosphere_hybrid_sigma_pressure_coordinate']
    assert standard_names['time'] == ['Time']
    assert standard_names['humidity_mixing_ratio'] == ['Q2', 'QVAPOR']
    assert standard_names['air_temperature'] == ['T2']

    # Check for time dimension reduction
    assert ds['z'].shape == (39,)
    assert ds['z_stag'].shape == (40,)
    assert ds['XLAT'].shape == ds['XLONG'].shape == (29, 31)
    assert ds['XLAT_U'].shape == ds['XLONG_U'].shape == (29, 32)
    assert ds['XLAT_V'].shape == ds['XLONG_V'].shape == (30, 31)

    # Check for diagnostic variable calculation
    assert 'air_potential_temperature' in ds.data_vars
    assert 'air_pressure' in ds.data_vars
    assert 'geopotential' in ds.data_vars
    assert 'geopotential_height' in ds.data_vars
    assert 'T' not in ds.data_vars
    assert 'P' not in ds.data_vars
    assert 'PB' not in ds.data_vars
    assert 'PH' not in ds.data_vars
    assert 'PHB' not in ds.data_vars
